﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.IO;
using BioLab.ImageProcessing;
using BioLab.Common;
using BioLab.Math.LinearAlgebra;
using BioLab.Biometrics.Face;
using BioLab.DimensionalityReduction;
using BioLab.GUI.DataViewers;

namespace FaceRecognition
{
    public partial class EigenfaceForm : Form
    {
        private EigenfaceLeftEnroller leftEnroller;
        private EigenfaceRightEnroller rightEnroller;
        private EigenfaceMatcher matcher;
        private Vector[] trainingImageFeatures;
        private List<string> imageNames = new List<string>();
        private int width, height;
        private EigenfaceTemplate[] templates;
        private int[] templatesID;
        private ImageViewer[] eigenfaces;
        private ImageViewer[] templPrototypes;
        private int numEigenfaces = 10;

        public EigenfaceForm()
        {
            InitializeComponent();
            CreateImageViewers();
            matcher = new EigenfaceMatcher();
        }

        private void CreateImageViewers()
        {
            eigenfaces = new ImageViewer[numEigenfaces];
            int i = 0;

            for (int r = 0; r < 2; r++)
            {
                for (int c = 0; c < numEigenfaces / 2; c++)
                {
                    eigenfaces[i] = new ImageViewer();
                    tableLayoutPanelEigenfaces.Controls.Add(eigenfaces[i++], c, r); 
                }
            }
            templPrototypes = new ImageViewer[numEigenfaces / 2];
            for (int j = 0; j < numEigenfaces / 2; j++)
            {
                templPrototypes[j] = new ImageViewer();
                tableLayoutPanelNTempl.Controls.Add(templPrototypes[j], j, 0);
            }
        }

        private void buttonBrowse_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "Text files (*.txt)|*.txt|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK && dlg.CheckFileExists)
            {
                textBoxTrainingFile.Text = dlg.FileName;
            }
        }

        private void buttonTraining_Click(object sender, EventArgs e)
        {
            int numSubjects, numImages;
            List<Image<byte>> imageSet = new List<Image<byte>>();
            List<int> subjectIDs = new List<int>();

            if (textBoxTrainingFile.Text != "")
            {
                string imagePath = textBoxTrainingFile.Text.Substring(0, textBoxTrainingFile.Text.LastIndexOf('\\'));
                string[] fileContent = File.ReadAllLines(textBoxTrainingFile.Text);
                if (fileContent.Length > 0)
                {
                    ResetEigenfaces();
                    Cursor.Current = Cursors.WaitCursor;
                    string[] row = fileContent[0].Split(' ', '\t');
                    numSubjects = Convert.ToInt16(row[0]);
                    string imageName;
                    if (numSubjects > 0)
                    {
                        templates = new EigenfaceTemplate[numSubjects];
                        for (int i = 1; i < fileContent.Length; i++)
                        {
                            row = fileContent[i].Split(' ', '\t');
                            int id = Convert.ToInt16(row[0]);
                            numImages = Convert.ToInt16(row[1]);
                            imageName = imagePath + "\\" + row[2];
                            imageNames.Add(imageName);
                            for (int j = 0; j < numImages; j++)
                            {
                                imageSet.Add(ImageBase.LoadFromFile(imageName).ToByteImage());
                                subjectIDs.Add(id);
                            }
                        }
                        if (imageSet.Count > 0)
                        {
                            Training(imageSet, subjectIDs, numSubjects);
                            width = imageSet[0].Width;
                            height = imageSet[0].Height;
                            DrawEigenFaces(leftEnroller.KL);
                        }
                    }
                    Cursor.Current = Cursors.Default;
                }
            }
        }

        private void ResetEigenfaces()
        {
            for (int i = 0; i < numEigenfaces; i++)
            {
                eigenfaces[i].Image = null;
            }
        }

        private void Training(List<Image<byte>> imageSet, List<int> subjectIDs, int numSujects)
        {
            int dimK = Convert.ToInt32(textBoxDimK.Text);

            trainingImageFeatures = new Vector[imageSet.Count];
            FaceFeatureExtractor featureExtractor = new PixelIntensityFeatureExtractor();
            leftEnroller = EigenfaceLeftEnroller.Train(imageSet, featureExtractor, dimK);
            rightEnroller = new EigenfaceRightEnroller(leftEnroller.KL, featureExtractor);

            int j = 0;
            FaceSet faceSet;
            templatesID = new int[numSujects];
            for (int i = 0; i < numSujects; i++)
            {
                faceSet = new FaceSet();
                int currID = subjectIDs[j];
                while (j < subjectIDs.Count && subjectIDs[j] == currID)
                {
                    faceSet.Add(imageSet[j++]);
                }
                templates[i] = leftEnroller.Enroll(faceSet);
                templatesID[i] = currID; 
            }
        }

        private void DrawEigenFaces(KLTransform kl)
        {
            int dimK = kl.ReducedDimensionality;
            Image<byte> eigenFace = new Image<byte>(width, height);
            int imgPerRow = numEigenfaces / 2;

            for (int i = 0; i < Math.Min(dimK, imgPerRow); i++)
            {
                eigenfaces[i].Image = ConvertToImage(kl.Bases.GetRow(i));
            }
            int diff = Math.Min(dimK - imgPerRow, imgPerRow);
            for (int i = 0; i < diff; i++)
            {
                eigenfaces[i + imgPerRow].Image = ConvertToImage(kl.Bases.GetRow(kl.Eigenvalues.Count - diff + i));
            }
        }

        private Image<byte> ConvertToImage(Vector v)
        {
            Image<double> image = new Image<double>(width, height);
            for (int i = 0; i < v.Count; i++)
            {
                image[i] = v[i];
            }
            return image.ToByteImage(PixelConversionMethod.Stretch);
        }

        private void buttonIdentify_Click(object sender, EventArgs e)
        {
            if (leftEnroller != null && rightEnroller != null && matcher != null && imageViewerInputImage.Image!=null)
            {
                int[] ids = new int[5];
                ids = Identify();
                for (int i = 0; i < numEigenfaces / 2; i++)
                {
                    templPrototypes[i].Image = ImageBase.LoadFromFile(imageNames[ids[i] - 1]);
                }
                FeatureVector reduced = leftEnroller.KL.Reduce(new PixelIntensityFeatureExtractor().ExtractFeatures(imageViewerInputImage.Image.ToByteImage()));
                imageViewerApprox.Image = ConvertToImage(leftEnroller.KL.BackProjection(reduced).Features);
            }
        }

        private int[] Identify()
        {
            int[] ids = new int[5];

            Vector f = rightEnroller.Enroll(imageViewerInputImage.Image.ToByteImage());
            Vector scores = new Vector(templates.Length);
            int[] idx = new int[templates.Length];
            for (int i = 0; i < templates.Length; i++)
            {
                scores[i] = matcher.Match(templates[i], f);
                idx[i] = templatesID[i];
            }
            Array.Sort(scores.ToArray(), idx);
            for (int i = 0; i < 5; i++)
		    {
                ids[i] = idx[scores.Count - i - 1];
		    }
            return ids;
        }

        private void fileListControlInputImage_SelectedFileNameChanged(object sender, EventArgs e)
        {
            imageViewerInputImage.Image = ImageBase.LoadFromFile(fileListControlInputImage.SelectedFileName);
            imageViewerApprox.Image = null;
            for (int i = 0; i < templPrototypes.Length; i++)
            {
                templPrototypes[i].Image = null;
            }
        }

        private void buttonSave_Click(object sender, EventArgs e)
        {
            if (leftEnroller != null && leftEnroller.KL != null)
            {
                FolderBrowserDialog dlg = new FolderBrowserDialog();
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    leftEnroller.SaveTrainingToFile(dlg.SelectedPath + "\\Eigenspace.kl");
                    SaveTemplates(dlg.SelectedPath);
                    List<string> info = imageNames;
                    info.Add(string.Format("{0}\t{1}", width, height));
                    File.WriteAllLines(dlg.SelectedPath + "\\prototypes.txt", info.ToArray());
                }
            }
        }

        private void SaveTemplates(string path)
        {
            for (int i = 0; i < templates.Length; i++)
            {
                templates[i].SaveToFile(path + "\\" + templatesID[i].ToString() + ".dat");
            }
        }
        
        private void buttonLoad_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                textBoxTrainingFile.Text = "";
                leftEnroller = EigenfaceLeftEnroller.FromFile(dlg.SelectedPath + "\\Eigenspace.kl", new PixelIntensityFeatureExtractor());
                rightEnroller = new EigenfaceRightEnroller(leftEnroller.KL, leftEnroller.FeatureExtractor);
                string[] templFiles = Directory.GetFiles(dlg.SelectedPath, "*.dat");
                templates = new EigenfaceTemplate[templFiles.Length];
                templatesID = new int[templFiles.Length];
                for (int i = 0; i < templFiles.Length; i++)
                {
                    templates[i] = EigenfaceTemplate.LoadFromFile(templFiles[i]);
                    templatesID[i] = Convert.ToInt32(templFiles[i].Substring(templFiles[i].LastIndexOf("\\")+1).Replace(".dat",""));
                }
                string[] prototypes = File.ReadAllLines(dlg.SelectedPath + "\\prototypes.txt");
                imageNames = new List<string>();
                for (int i = 0; i < prototypes.Length - 1; i++)
                {
                    imageNames.Add(prototypes[i]);
                }
                string[] temp = prototypes[prototypes.Length - 1].Split('\t');
                width = Convert.ToInt32(temp[0]);
                height = Convert.ToInt32(temp[1]);
                textBoxDimK.Text = leftEnroller.KL.ReducedDimensionality.ToString();
                DrawEigenFaces(leftEnroller.KL);
            }
        }
    }
}
